#include "vtkapp.h"
#include "fluidgui.h"

extern MyVTKApp** gppVTKApp;
#define APP (*gppVTKApp)
#define APP_POINTER gppVTKApp;

void IsoSurfaceMode(Fl_Round_Button* btn, void* data){

	if(APP == NULL)
		return;

	RayCastRadio->value(0);
	InterpolationChoice->deactivate();
	VisualizationChoice->deactivate();
	SampleDistanceInput->deactivate();
	UpdateRayCastSettingBtn->deactivate();

	VTKAppMessage *msg = new VTKAppMessage();
	msg->msg = MyVTKApp::VTKAPP_ISOSURFACE_MODE;
	msg->msgData = NULL;

	APP->Message(msg);

	msg = new VTKAppMessage();
	msg->msg = MyVTKApp::VTKAPP_RENDER;
	msg->msgData = NULL;

	APP->Message(msg);

	msg = NULL;
}


void RayCastMode(Fl_Round_Button* btn, void* data){

	if(APP == NULL)
		return;

	IsoSurfaceRadio->value(0);
	InterpolationChoice->activate();
	VisualizationChoice->activate();
	SampleDistanceInput->activate();
	UpdateRayCastSettingBtn->activate();

	VTKAppMessage *msg = new VTKAppMessage();
	msg->msg = MyVTKApp::VTKAPP_RAYCAST_MODE;
	msg->msgData = NULL;

	APP->Message(msg);

	msg = new VTKAppMessage();
	msg->msg = MyVTKApp::VTKAPP_RENDER;
	msg->msgData = NULL;

	APP->Message(msg);

	msg = NULL;
}

void IsolateFeet(Fl_Button* b, void* d){
	
	if(APP == NULL)
		return;

	VTKAppMessage *msg = new VTKAppMessage();
	msg->msg = MyVTKApp::VTKAPP_ISOLATE_FEET;
	msg->msgData = NULL;

	APP->Message(msg);

	msg = new VTKAppMessage();
	msg->msg = MyVTKApp::VTKAPP_RENDER;
	msg->msgData = NULL;

	APP->Message(msg);

	msg = NULL;
	
}

void IsolateHead(Fl_Button* b, void* d){
	
	if(APP == NULL)
		return;

	VTKAppMessage *msg = new VTKAppMessage();
	msg->msg = MyVTKApp::VTKAPP_ISOLATE_HEAD;
	msg->msgData = NULL;

	APP->Message(msg);

	msg = new VTKAppMessage();
	msg->msg = MyVTKApp::VTKAPP_RENDER;
	msg->msgData = NULL;

	APP->Message(msg);

	msg = NULL;
}


void DataSetResolution(Fl_Choice* ch, void* d){
		
	if(APP == NULL)
		return;

	VTKAppMessage *msg = new VTKAppMessage();
	msg->msg = MyVTKApp::VTKAPP_SET_DATA_RESOLUTION;
	int value = ch->value() + 1;
	char* buf = new char[sizeof(int) + 1];

	memcpy(buf,&value,sizeof(int));
	msg->msgData = (void*)buf;
	buf = NULL;

	APP->Message(msg);

	msg = new VTKAppMessage();
	msg->msg = MyVTKApp::VTKAPP_RENDER;
	msg->msgData = NULL;

	APP->Message(msg);

	msg = NULL;
}

void AdjustSkinTransparencyCb(Fl_Value_Slider* sl, void* d){

	if(APP == NULL)
		return;

	VTKAppMessage *msg = new VTKAppMessage();
	msg->msg = MyVTKApp::VTKAPP_SET_SKIN_TRANSPARENCY;
	double value = sl->value();

	char* buf = new char[sizeof(double)];
	memcpy(buf,&value,sizeof(double));
	msg->msgData = (void*)buf;
	buf = NULL;

	APP->Message(msg);

	msg = new VTKAppMessage();
	msg->msg = MyVTKApp::VTKAPP_RENDER;
	msg->msgData = NULL;

	APP->Message(msg);

	msg = NULL;
}

void UpdateRayCastSettingCb(Fl_Button* b, void* d){

	if(APP == NULL)
		return;

	static double lastSampDist = 1.0;
	static int visChoice = 0;
	static int interpChoice = 0;

	VTKAppMessage *msg = NULL;
	char* buf = NULL;
	//if there were any changes then send a message
	if(lastSampDist != SampleDistanceInput->value())
	{
		lastSampDist = SampleDistanceInput->value();
		msg = new VTKAppMessage();
		msg->msg = MyVTKApp::VTKAPP_SET_SAMPLE_DISTANCE;
		buf = new char[ sizeof(double) ];
		memcpy(buf,&lastSampDist,sizeof(double));
		msg->msgData = (void*)buf;
		buf = NULL;
		APP->Message(msg);
		msg = NULL;
		
	}

	if(visChoice != VisualizationChoice->value())
	{
		visChoice = VisualizationChoice->value();
		msg = new VTKAppMessage();
		switch(visChoice)
		{
		case 0:
			msg->msg = MyVTKApp::VTKAPP_MIP_RAYCAST;
			msg->msgData = NULL;
			break;
		case 1:
			msg->msg = MyVTKApp::VTKAPP_COMPOSITE_RAYCAST;
			msg->msgData = NULL;
			break;
		default:
			msg->msg = MyVTKApp::VTKAPP_MIP_RAYCAST;
			msg->msgData = NULL;
			break;
		}
		buf = NULL;
		APP->Message(msg);
		msg = NULL;
	}
	
	if(interpChoice != InterpolationChoice->value())
	{
		interpChoice = InterpolationChoice->value();
				msg = new VTKAppMessage();
		switch(visChoice)
		{
		case 0:
			msg->msg = MyVTKApp::VTKAPP_ZERO_ORDER_INTERPOLATION;
			msg->msgData = NULL;
			break;
		case 1:
			msg->msg = MyVTKApp::VTKAPP_LINEAR_INTERPOLATION;
			msg->msgData = NULL;
			break;
		default:
			msg->msg = MyVTKApp::VTKAPP_ZERO_ORDER_INTERPOLATION;
			msg->msgData = NULL;
			break;
		}
		buf = NULL;
		APP->Message(msg);
		msg = NULL;
	}

	//send a render message no matter what...doesn't really hurt =)
	msg = new VTKAppMessage();
	msg->msg = MyVTKApp::VTKAPP_RENDER;
	msg->msgData = NULL;

	APP->Message(msg);

	msg = NULL;

}

void UpdateGeneralCb(Fl_Button* btn, void* d)
{
	AdjustSkinTransparencyCb(SkinTransparencySlider,NULL);
	DataSetResolution(ResolutionChoice,NULL);
}
