#include "vtkapp.h"
#include <stdlib.h>
#include <stdio.h>
#include <conio.h>

#define X 0
#define Y 1
#define Z 2
#define W 3

#define FOOT 0
#define HEAD 1
#define BONE 0
#define SKIN 1
#define RAYFOOT 2
#define RAYHEAD 3

#define RED 0
#define GREEN 1
#define BLUE 2
#define YELLOW 3
#define ALPHA 3

///-----------------------------------
// VTK App messages supported

const int MyVTKApp::VTKAPP_RENDER = 1;
const int MyVTKApp::VTKAPP_ISOSURFACE_MODE = 2;
const int MyVTKApp::VTKAPP_RAYCAST_MODE = 3;
const int MyVTKApp::VTKAPP_SET_DATA_RESOLUTION = 4;
const int MyVTKApp::VTKAPP_SET_SKIN_TRANSPARENCY = 5;
const int MyVTKApp::VTKAPP_ISOLATE_FEET = 6;
const int MyVTKApp::VTKAPP_ISOLATE_HEAD = 7;
const int MyVTKApp::VTKAPP_LINEAR_INTERPOLATION = 8;
const int MyVTKApp::VTKAPP_ZERO_ORDER_INTERPOLATION = 9;
const int MyVTKApp::VTKAPP_SET_SAMPLE_DISTANCE = 10;
const int MyVTKApp::VTKAPP_COMPOSITE_RAYCAST = 11;
const int MyVTKApp::VTKAPP_MIP_RAYCAST = 12;

//--------------------------------------------------
//Private constants
const int MyVTKApp::FOOT_BONE = 0;
const int MyVTKApp::FOOT_SKIN = 1;
const int MyVTKApp::HEAD_BONE = 2;
const int MyVTKApp::HEAD_SKIN = 3;

//---------------------------------------
MyVTKApp::MyVTKApp(HWND& hwnd, int width, int height):head(NULL),tail(NULL)
{
	parentWindow = hwnd;
	renderWindow = vtkRenderWindow::New();
	renderWindow->SetSize(width,height);
	renderWindow->SetParentId(hwnd);

	render1 = new vtkOpenGLRenderer*[3];
	for(int k = 0; k < 3; k++)
	{
		render1[k] = vtkOpenGLRenderer::New();
		render1[k]->SetBackground(0,0,0);
	}

	interactor = vtkRenderWindowInteractor::New();

	renderWindow->AddRenderer(render1[0]);
	interactor->SetRenderWindow(renderWindow);

	//------------------------------------------
	//Open the RAW Visible Woman Data & generate clippings
	dataSetReader = new vtkImageReader2*[3];
	for(k = 0; k < 3; k++)
	{
		dataSetReader[k] = vtkImageReader2::New();
		dataSetReader[k]->SetFileName("Female.raw");
		dataSetReader[k]->SetDataScalarType(VTK_UNSIGNED_SHORT);
		dataSetReader[k]->SetDataByteOrder(0);
		dataSetReader[k]->SetNumberOfScalarComponents(1);
		dataSetReader[k]->SetFileDimensionality(3);
		dataSetReader[k]->SetDataSpacing(1,1,1);
		dataSetReader[k]->SetDataOrigin(0,0,0);
		dataSetReader[k]->SetDataExtent(0,255,0,255,0,576);
	}

	imageClippings = new vtkExtractVOI*[4];
	imageClippings[FOOT] = vtkExtractVOI::New();
	imageClippings[HEAD] = vtkExtractVOI::New();
	imageClippings[RAYFOOT] = vtkExtractVOI::New();
	imageClippings[RAYHEAD] = vtkExtractVOI::New();	

	imageClippings[FOOT]->SetInput( dataSetReader[0]->GetOutput() );	//foot clipping first
	imageClippings[FOOT]->SetVOI( 64, 180, 64, 200, 501, 576 );
	imageClippings[HEAD]->SetInput( dataSetReader[0]->GetOutput() );	//head clipping next
	imageClippings[HEAD]->SetVOI( 64 ,185 ,64 ,200 ,0 ,90 );
	imageClippings[RAYFOOT]->SetInput( dataSetReader[1]->GetOutput() );	//foot clipping first
	imageClippings[RAYFOOT]->SetVOI( 64, 180, 64, 200, 501, 576 );
	imageClippings[RAYHEAD]->SetInput( dataSetReader[2]->GetOutput() );	//head clipping next
	imageClippings[RAYHEAD]->SetVOI( 64 ,185 ,64 ,200 ,0 ,90 );

	//create marching cube contour filters
	contours = new vtkMarchingContourFilter*[4];
	lookup = new vtkLookupTable*[2];
	lookup[FOOT] = vtkLookupTable::New();
	lookup[HEAD] = vtkLookupTable::New();

	lookup[BONE]->SetNumberOfTableValues(1);
	lookup[BONE]->SetTableValue(0, 1, .5, 0);
	lookup[SKIN]->SetNumberOfTableValues(1);
	lookup[SKIN]->SetTableValue(0, 0, 1, 1);

//	contours = new vtkContourFilter*[4];
	for(int i = 0; i < 4; i++)
	{
		contours[i] = vtkMarchingContourFilter::New();
		//contours[i] = vtkContourFilter::New();
		contours[i]->SetNumberOfContours(1);
	}

	contours[MyVTKApp::FOOT_BONE]->SetInput( (vtkDataSet*) imageClippings[FOOT]->GetOutput() );
	contours[MyVTKApp::FOOT_SKIN]->SetInput( (vtkDataSet*) imageClippings[FOOT]->GetOutput() );
	contours[MyVTKApp::HEAD_BONE]->SetInput( (vtkDataSet*) imageClippings[HEAD]->GetOutput() );
	contours[MyVTKApp::HEAD_SKIN]->SetInput( (vtkDataSet*) imageClippings[HEAD]->GetOutput() );

	contours[MyVTKApp::FOOT_BONE]->SetValue(0, 1150.0);		//Set the bone contour values
	contours[MyVTKApp::HEAD_BONE]->SetValue(0, 1150.0);
	contours[MyVTKApp::FOOT_SKIN]->SetValue(0, 450);		//set the skin contour values
	contours[MyVTKApp::HEAD_SKIN]->SetValue(0, 450);

	//create the geometry filters
	geoFilters = new vtkGeometryFilter*[4];
	//crate the mappers
	polyMappers = new vtkPolyDataMapper*[4];
	//create the actors
	isoActors = new vtkLODActor*[4];

	for(i = 0; i < 4; i++)
	{
		geoFilters[i] = vtkGeometryFilter::New();
		polyMappers[i] = vtkPolyDataMapper::New();
		isoActors[i] = vtkLODActor::New();
		isoActors[i]->SetNumberOfCloudPoints( 1000 );

		geoFilters[i]->SetInput( (vtkDataSet*) contours[i]->GetOutput() );
		polyMappers[i]->SetInput( (vtkPolyData*) geoFilters[i]->GetOutput() );
		isoActors[i]->SetMapper( polyMappers[i] );
		isoActors[i]->GetProperty()->SetRepresentationToSurface();
		isoActors[i]->GetProperty()->SetInterpolationToGouraud();

		render1[0]->AddActor( isoActors[i] );

	}
	
	polyMappers[MyVTKApp::FOOT_BONE]->SetLookupTable( lookup[BONE] );
	polyMappers[MyVTKApp::HEAD_BONE]->SetLookupTable( lookup[BONE] );
	polyMappers[MyVTKApp::FOOT_SKIN]->SetLookupTable( lookup[SKIN] );
	polyMappers[MyVTKApp::HEAD_SKIN]->SetLookupTable( lookup[SKIN] );

	//---------------------------
	//Setup the Volume Renderings
	opacityFunc = vtkPiecewiseFunction::New();
	opacityFunc->AddSegment(0, 0, 449, 0);
	opacityFunc->AddSegment( 450, 1, 1149, 1);
	opacityFunc->AddSegment( 1150, 1, 65535, 1 );


	colorFunc = vtkColorTransferFunction::New();
	colorFunc->AddRGBPoint( 1150, .5, 1, 0);
	colorFunc->AddRGBPoint( 450, 0, 0, 1);
	//colorFunc->Build();
	
	compositeRay = vtkVolumeRayCastCompositeFunction::New();
	compositeRay->SetCompositeMethodToInterpolateFirst();
	mipRay = vtkVolumeRayCastMIPFunction::New();
	mipRay->SetMaximizeMethodToScalarValue();

	volProperties = new vtkVolumeProperty*[2];
	volMappers = new vtkVolumeRayCastMapper*[2];
	volumes = new vtkVolume*[2];
	
	for( i = 0; i < 2; i++)
	{
		volProperties[i] = vtkVolumeProperty::New();
		volMappers[i] = vtkVolumeRayCastMapper::New();
		volumes[i] = vtkVolume::New();
		
		volProperties[i]->SetColor( colorFunc );
		volProperties[i]->SetScalarOpacity( opacityFunc );
		volProperties[i]->ShadeOn();
		volProperties[i]->SetInterpolationTypeToLinear();		//default interpolation is linear

		volMappers[i]->SetVolumeRayCastFunction( mipRay );		//MIP ray cast is default
		volMappers[i]->SetInput( imageClippings[i + 2]->GetOutput() );
		volMappers[i]->CroppingOff();

		volumes[i]->SetMapper( volMappers[i] );
		volumes[i]->SetProperty( volProperties[i] );
		volumes[i]->SetPosition( 0, 0, 0);

	}
    // don't add the volumes to the renderer
	fullView = true;
	headView = false;
	rayCast = false;
	render1[1]->AddVolume( volumes[HEAD] );
	render1[2]->AddVolume( volumes[FOOT] );
	for(k = 0; k < 3; k++)
		render1[k]->ResetCamera();

	
}

MyVTKApp::~MyVTKApp()
{
	interactor->Delete();
	renderWindow->Delete();
	render1[0]->Delete();
	render1[1]->Delete();
	render1[2]->Delete();
	delete [] render1;
	for(int k = 0; k < 3; k++)
		dataSetReader[k]->Delete();
	delete [] dataSetReader;
	imageClippings[0]->Delete();
	imageClippings[1]->Delete();
	imageClippings[2]->Delete();
	imageClippings[3]->Delete();
	delete [] imageClippings;
	for(int i = 0; i < 4; i++)
	{
		isoActors[i]->Delete();
		polyMappers[i]->Delete();
		geoFilters[i]->Delete();
		contours[i]->Delete();

	}

	delete [] isoActors;
	delete [] polyMappers;
	delete [] geoFilters;
	delete [] contours;

	
}

void MyVTKApp::Render()
{

	//make sure that if we are in raycast mode that the number of visible volumes == 1
/*	if(rayCast)
	{
		if(render1->VisibleVolumeCount() < 1)
			render1->ResetCameraClippingRange();

		//MessageBox(this->parentWindow, "WARNING", "LOST THE VOLUME!", MB_OK);
	}
*/
	if(renderWindow)
		renderWindow->Render();

}

void MyVTKApp::Message(VTKAppMessage* vtkAppMsg)
{
	VTKAppMessageQueue* newNode = new VTKAppMessageQueue();
	newNode->next = NULL;
	newNode->pMsgStruct = vtkAppMsg;
	//MessageBox(parentWindow,"Message Recieved","Acknowledgement",MB_OK);
	if(!head)
	{
		head = newNode;
		tail = newNode;
	}
	else
	{
		tail->next = newNode;
		tail = newNode;
		tail->next = NULL;
	}
}

void MyVTKApp::ProcessMessages()
{
	while(head)
	{
		VTKAppMessageQueue* temp;
		float *sampDist = NULL;
		float *posTemp = NULL;
		float *tempFloatArray = NULL;
		double *tempDoubleArray = NULL;

		int set;
		switch(head->pMsgStruct->msg)
		{
		case MyVTKApp::VTKAPP_RENDER:					//////////////////////////////////
			this->Render();
			break;
		case MyVTKApp::VTKAPP_ISOLATE_FEET:
			//render1->ResetCamera();
			//MessageBox(this->parentWindow, "ISOLATING FEET", "Action", MB_OK);
			fullView = false;
			headView = false;
			isoActors[MyVTKApp::HEAD_BONE]->SetVisibility(0);
			isoActors[MyVTKApp::HEAD_SKIN]->SetVisibility(0);
			isoActors[MyVTKApp::FOOT_BONE]->SetVisibility(1);
			isoActors[MyVTKApp::FOOT_SKIN]->SetVisibility(1);
			//volumes[FOOT]->SetVisibility(1);
			//volumes[HEAD]->SetVisibility(0);

			if(rayCast)
			{
				//render1->ResetCamera( volumes[FOOT]->GetBounds() );
				render1[2]->GetActiveCamera()->SetClippingRange(0, 2000); //( volumes[FOOT]->GetBounds() );
				posTemp = volumes[FOOT]->GetPosition();
				posTemp[2] -= 500;
				render1[2]->GetActiveCamera()->SetPosition( posTemp[0], posTemp[1], posTemp[2] );
				posTemp[2] += 500;
				//render1->GetActiveCamera()->SetViewPlaneNormal( -1 * posTemp[0], -1 * posTemp[1], -1 * posTemp[2] );
				render1[2]->GetActiveCamera()->SetFocalPoint( posTemp[0], posTemp[1], posTemp[2] );
				renderWindow->RemoveRenderer(render1[0]);
				renderWindow->RemoveRenderer(render1[1]);
				renderWindow->AddRenderer(render1[2]);
				//render1[2]->ResetCamera();
			}
			else
			{
				renderWindow->RemoveRenderer(render1[1]);
				renderWindow->RemoveRenderer(render1[2]);
				renderWindow->AddRenderer( render1[0] );
				render1[0]->ResetCamera();
			}
			
			break;
		case MyVTKApp::VTKAPP_ISOLATE_HEAD:
			//render1->ResetCamera();
			//MessageBox(this->parentWindow, "ISOLATING HEAD", "Action", MB_OK);
			fullView = false;
			headView = true;
			isoActors[MyVTKApp::FOOT_BONE]->SetVisibility(0);
			isoActors[MyVTKApp::FOOT_SKIN]->SetVisibility(0);
			isoActors[MyVTKApp::HEAD_BONE]->SetVisibility(1);
			isoActors[MyVTKApp::HEAD_SKIN]->SetVisibility(1);
			//volumes[FOOT]->SetVisibility(0);

			if(rayCast)
			{
				//render1->ResetCamera( volumes[HEAD]->GetBounds() );
				//render1->ResetCameraClippingRange( volumes[HEAD]->GetBounds() );
			/*	render1->GetActiveCamera()->SetClippingRange(0, 2000);
				posTemp = volumes[HEAD]->GetPosition();
				posTemp[2] -= 500;
				render1->GetActiveCamera()->SetPosition( posTemp[0], posTemp[1], posTemp[2] );
				posTemp[2] += 500;
				//render1->GetActiveCamera()->SetViewPlaneNormal( -1 * posTemp[0], -1 * posTemp[1], -1 * posTemp[2] );
				render1->GetActiveCamera()->SetFocalPoint( posTemp[0], posTemp[1], posTemp[2] );*/
				renderWindow->RemoveRenderer(render1[0]);
				renderWindow->RemoveRenderer(render1[2]);
				renderWindow->AddRenderer(render1[1]);
			}
			else
			{
				renderWindow->RemoveRenderer(render1[1]);
				renderWindow->RemoveRenderer(render1[2]);
				renderWindow->AddRenderer(render1[0]);
				render1[0]->ResetCamera();
			}
			
			break;
		case MyVTKApp::VTKAPP_ISOSURFACE_MODE:
			//render1[0]->AddActor( isoActors[MyVTKApp::HEAD_BONE] );
			//render1[0]->AddActor( isoActors[MyVTKApp::FOOT_BONE] );
			//render1[0]->AddActor( isoActors[MyVTKApp::HEAD_SKIN] );
			//render1[0]->AddActor( isoActors[MyVTKApp::FOOT_SKIN] );
			
			if( fullView )
			{
				isoActors[MyVTKApp::FOOT_BONE]->SetVisibility(1);
				isoActors[MyVTKApp::FOOT_SKIN]->SetVisibility(1);
				isoActors[MyVTKApp::HEAD_BONE]->SetVisibility(1);
				isoActors[MyVTKApp::HEAD_SKIN]->SetVisibility(1);
				
			}
			else if( !headView)
			{	
				isoActors[MyVTKApp::FOOT_BONE]->SetVisibility(1);
				isoActors[MyVTKApp::FOOT_SKIN]->SetVisibility(1);
				isoActors[MyVTKApp::HEAD_BONE]->SetVisibility(0);
				isoActors[MyVTKApp::HEAD_SKIN]->SetVisibility(0);		
			}
			else
			{
				isoActors[MyVTKApp::FOOT_BONE]->SetVisibility(0);
				isoActors[MyVTKApp::FOOT_SKIN]->SetVisibility(0);
				isoActors[MyVTKApp::HEAD_BONE]->SetVisibility(1);
				isoActors[MyVTKApp::HEAD_SKIN]->SetVisibility(1);

			}
			rayCast = false;
			renderWindow->RemoveRenderer(render1[1]);
			renderWindow->RemoveRenderer(render1[2]);
			renderWindow->AddRenderer(render1[0]);
			
			render1[0]->ResetCamera();


			break;
		case MyVTKApp::VTKAPP_RAYCAST_MODE:
			//find out which set of data was visible
			//force to be HEAD as default raycast
			//render1->AddVolume( volumes[HEAD] );
			//render1->RemoveVolume( volumes[FOOT] );
			//render1->AddVolume( volumes[HEAD] );
			headView = true;
			fullView = false;
			rayCast = true;
			//volumes[HEAD]->SetVisibility(1);
			//volumes[FOOT]->SetVisibility(0);
			renderWindow->RemoveRenderer(render1[0]);
			renderWindow->RemoveRenderer(render1[2]);
			renderWindow->AddRenderer(render1[1]);

			//render1->RemoveActor( isoActors[MyVTKApp::HEAD_BONE] );
			//render1->RemoveActor( isoActors[MyVTKApp::FOOT_BONE] );
			//render1->RemoveActor( isoActors[MyVTKApp::HEAD_SKIN] );
			//render1->RemoveActor( isoActors[MyVTKApp::FOOT_SKIN] );
            
			//render1->ResetCamera();
			//render1->ResetCameraClippingRange();
			//render1->ResetCamera();
			//volumes[FOOT]->Update();
			//render1->ResetCamera( volumes[HEAD]->GetBounds() );
			//render1->ResetCameraClippingRange( volumes[HEAD]->GetBounds() );
			//render1->GetActiveCamera()->SetClippingRange(0, 2000);
			//set the position to be the same as the center of the object
			/******************************************/
			/*
			posTemp = volumes[HEAD]->GetPosition();
			posTemp[2] -= 500;
			render1->GetActiveCamera()->SetPosition( posTemp[0], posTemp[1], posTemp[2] );
			posTemp[2] += 500;
			//render1->GetActiveCamera()->SetViewPlaneNormal( -1 * posTemp[0], -1 * posTemp[1], -1 * posTemp[2] );
			render1->GetActiveCamera()->SetFocalPoint( posTemp[0], posTemp[1], posTemp[2] );
			/***********************************************/
			//tempFloatArray = new float[12];
			//render1->GetActiveCamera()->GetPosition(tempFloatArray);
			//render1->GetActiveCamera()->GetFocalPoint(&tempFloatArray[3]);
			//posTemp = volumes[HEAD]->GetCenter();
			//memcpy( &tempFloatArray[9], posTemp, sizeof(float) * 3);

	//		tempFloatArray[6] = tempFloatArray[3] - tempFloatArray[0];
	//		tempFloatArray[7] = tempFloatArray[4] - tempFloatArray[1];
	//		tempFloatArray[8] = tempFloatArray[5] - tempFloatArray[2];
			
			//tempFloatArray[6] = tempFloatArray[3] - tempFloatArray[9];
			//tempFloatArray[7] = tempFloatArray[4] - tempFloatArray[10];
			//tempFloatArray[8] = tempFloatArray[5] - tempFloatArray[11];
	
			//posTemp = volumes[HEAD]->GetCenter();
			
			//posTemp[0] = posTemp[0] - tempFloatArray[6];
			//posTemp[1] = posTemp[1] - tempFloatArray[7];
			//posTemp[2] = posTemp[2] - tempFloatArray[8];

			//tempDoubleArray = render1->GetActiveCamera()->GetViewShear();

			//render1->GetActiveCamera()->SetPosition( posTemp );
			//render1->GetActiveCamera()->SetFocalPoint( tempFloatArray[9],
			//											tempFloatArray[10],
			//											tempFloatArray[11]);

			//delete tempFloatArray;
				
			//render1->GetActiveCamera()->SetViewShear(tempDoubleArray);
			break;
		case MyVTKApp::VTKAPP_SET_DATA_RESOLUTION:
			AdjustResolution(head->pMsgStruct->msgData);
			volumes[FOOT]->Update();
			volumes[HEAD]->Update();
			break;
		case MyVTKApp::VTKAPP_SET_SAMPLE_DISTANCE:
			volMappers[HEAD]->SetSampleDistance( *((float *) (head->pMsgStruct->msgData) ) );
			volMappers[FOOT]->SetSampleDistance( *((float *) (head->pMsgStruct->msgData) ) );
			volumes[FOOT]->Update();
			volumes[HEAD]->Update();
			break;
		case MyVTKApp::VTKAPP_SET_SKIN_TRANSPARENCY:
			SetSkinTransparency(head->pMsgStruct->msgData);
			opacityFunc->RemoveAllPoints();
			opacityFunc->AddSegment(0, 0, 449, 0);
			opacityFunc->AddSegment( 450, *((float *) (head->pMsgStruct->msgData) ), 1149, *((float *) (head->pMsgStruct->msgData) ) );
			//opacityFunc->AddSegment( 461, 0, 1149, 0 );
			opacityFunc->AddSegment( 1150, 1, 65535, 1);
			volProperties[0]->SetScalarOpacity( opacityFunc );
			volProperties[1]->SetScalarOpacity( opacityFunc );

			volumes[FOOT]->Update();
			volumes[HEAD]->Update();
			break;
		case MyVTKApp::VTKAPP_ZERO_ORDER_INTERPOLATION:
			volProperties[HEAD]->SetInterpolationTypeToNearest();
			volProperties[FOOT]->SetInterpolationTypeToNearest();
			volMappers[FOOT]->Update();
			volMappers[HEAD]->Update();
			volumes[FOOT]->Update();
			volumes[HEAD]->Update();
			break;
		case MyVTKApp::VTKAPP_LINEAR_INTERPOLATION:
			volProperties[HEAD]->SetInterpolationTypeToLinear();
			volProperties[FOOT]->SetInterpolationTypeToLinear();
			volMappers[FOOT]->Update();
			volMappers[HEAD]->Update();
			volumes[FOOT]->Update();
			volumes[HEAD]->Update();
			break;
		case MyVTKApp::VTKAPP_MIP_RAYCAST:
			volMappers[HEAD]->SetVolumeRayCastFunction( mipRay );
			volMappers[FOOT]->SetVolumeRayCastFunction( mipRay );
			volMappers[FOOT]->Update();
			volMappers[HEAD]->Update();
			volumes[FOOT]->Update();
			volumes[HEAD]->Update();
			break;
		case MyVTKApp::VTKAPP_COMPOSITE_RAYCAST:
			volMappers[HEAD]->SetVolumeRayCastFunction( compositeRay );
			volMappers[FOOT]->SetVolumeRayCastFunction( compositeRay );
			volMappers[FOOT]->Update();
			volMappers[HEAD]->Update();
			volumes[FOOT]->Update();
			volumes[HEAD]->Update();
			break;
		default:									//////////////////////////////////

			break;
		};
		
		temp = head;
		head = head->next;

		if(temp->pMsgStruct->msgData != NULL)
			delete [] temp->pMsgStruct->msgData;

		delete temp;
	}

}

bool MyVTKApp::HaveMessages()
{
	if( head != NULL)
		return true;
	else
		return false;
}

void MyVTKApp::SetSkinTransparency(void* d)
{
	double *alpha = (double*) d;
	//lookup[SKIN]->SetAlpha( (float) (*alpha) );
	isoActors[MyVTKApp::HEAD_SKIN]->GetProperty()->SetOpacity((float) (*alpha));
	isoActors[MyVTKApp::FOOT_SKIN]->GetProperty()->SetOpacity((float) (*alpha));

}

void MyVTKApp::AdjustResolution(void* d)
{
	int *res = (int*) d;
	imageClippings[0]->SetSampleRate(*res,*res,*res);
	imageClippings[1]->SetSampleRate(*res,*res,*res);
	imageClippings[2]->SetSampleRate(*res,*res,*res);
	imageClippings[3]->SetSampleRate(*res,*res,*res);
//	imageClippings[0]->Update();
//	imageClippings[1]->Update();
//	imageClippings[2]->Update();
//	imageClippings[3]->Update();
}
