#ifndef VTK_APP_H
#define VTK_APP_H


#include <vtk/vtkRenderWindow.h>
#include <vtk/vtkRenderer.h>
#include <vtk/vtkOpenGLRenderer.h>
#include <vtk/vtkRenderWindowInteractor.h>
#include <vtk/vtkConeSource.h>
#include <vtk/vtkPolyDataMapper.h>
#include <vtk/vtkActor.h>
#include <vtk/vtkProperty.h>
#include <vtk/vtkCamera.h>

//--------------------------------------
#include <vtk/vtkImageReader2.h>
#include <vtk/vtkLODActor.h>
#include <vtk/vtkLookupTable.h>
#include <vtk/vtkContourFilter.h>
#include <vtk/vtkImageClip.h>
#include <vtk/vtkGeometryFilter.h>
#include <vtk/vtkMarchingContourFilter.h>
#include <vtk/vtkExtractVOI.h>

//---------------------------------------
#include <vtk/vtkPiecewiseFunction.h>
#include <vtk/vtkColorTransferFunction.h>
#include <vtk/vtkVolumeProperty.h>
#include <vtk/vtkVolumeRayCastCompositeFunction.h>
#include <vtk/vtkVolumeRayCastMIPFunction.h>
#include <vtk/vtkVolumeRayCastMapper.h>
#include <vtk/vtkVolume.h>

//--------------------------------------
struct ImageFileData
{
	char* fileName;
	int width;
	int height;
	int depth;
};

struct ContourData
{
	int color;
	int level;
	int transparency;
};

struct LookupTableData
{
	float hueMin;
	float hueMax;
	float satMin;
	float satMax;
	float rngMin;
	float rngMax;
};

struct VTKAppMessage
{
	int	msg;
	void* msgData;	//assume the data is always in array form!  Single values are from arrays of size 1!
};

struct VTKAppMessageQueue
{
	VTKAppMessage* pMsgStruct;
	VTKAppMessageQueue* next;
};

//		VTK APPLICATION CLASS
//		- Change the Render method to what ever way you want
//----------------------------------
class MyVTKApp 
{
public:
	MyVTKApp(HWND&, int, int);
	virtual ~MyVTKApp();

	virtual void	Render();
	void	Message(VTKAppMessage* );	//recieve messages
	void	ProcessMessages();			//process messages in message queue
	bool	HaveMessages();

	void ResetColorMapTerrain(void);
	///-----------------------------------
	// VTK App messages supported
	static const int VTKAPP_RENDER;
	static const int VTKAPP_ISOSURFACE_MODE;
	static const int VTKAPP_RAYCAST_MODE;
	static const int VTKAPP_SET_DATA_RESOLUTION;
	static const int VTKAPP_SET_SKIN_TRANSPARENCY;
	static const int VTKAPP_ISOLATE_FEET;
	static const int VTKAPP_ISOLATE_HEAD;
	static const int VTKAPP_ZERO_ORDER_INTERPOLATION;
	static const int VTKAPP_LINEAR_INTERPOLATION;
	static const int VTKAPP_SET_SAMPLE_DISTANCE;
	static const int VTKAPP_MIP_RAYCAST;
	static const int VTKAPP_COMPOSITE_RAYCAST;


private:

	bool	Updated();
	
	HWND parentWindow;
	int terrainMode;

	VTKAppMessageQueue *head;
	VTKAppMessageQueue *tail;

	vtkOpenGLRenderer		**render1;
	vtkRenderWindow	*renderWindow;
	vtkRenderWindowInteractor	*interactor;


	vtkImageReader2	**dataSetReader;

	static const int FOOT_SKIN;
	static const int FOOT_BONE;
	static const int HEAD_SKIN;
	static const int HEAD_BONE;
	
	unsigned int skinLevel;
	unsigned int bonelevel;

	vtkExtractVOI **imageClippings;	// 2 image clips for foot and head isolation
	vtkMarchingContourFilter **contours;
	vtkLookupTable **lookup;

	//vtkContourFilter **contours;
	vtkGeometryFilter	**geoFilters;
	vtkPolyDataMapper **polyMappers;
	vtkLODActor **isoActors;		//actors for the isosurfaces		

	//REFER TO SIMPLERAYCAST.TCL FOR OBJECT NECESSARY TO DO RAYCAST SECTION!!!!!
	vtkPiecewiseFunction* opacityFunc;
	vtkColorTransferFunction *colorFunc;
	vtkVolumeProperty **volProperties;
	vtkVolumeRayCastCompositeFunction *compositeRay;
	vtkVolumeRayCastMIPFunction *mipRay;
	vtkVolumeRayCastMapper **volMappers;
	vtkVolume **volumes;

	bool headView;
	bool fullView;
	bool rayCast;

//	PRIVATE METHODS
	void SetSkinTransparency(void* );
	void AdjustResolution(void*);

};






#endif